import dataclasses
from typing import Optional

from qua.info import QuaMachineInfo


@dataclasses.dataclass
class ServerCapabilities:
    has_job_streaming_state: bool
    supports_multiple_inputs_for_element: bool
    supports_analog_delay: bool
    supports_shared_oscillators: bool
    supports_channel_weights: bool

    @staticmethod
    def build(qop_version: str, qua_implementation: Optional[QuaMachineInfo]):
        caps = (
            qua_implementation.capabilities
            if qua_implementation is not None
            else list()
        )
        return ServerCapabilities(
            has_job_streaming_state=_has_job_streaming_state(
                qop_version, qua_implementation
            ),
            supports_multiple_inputs_for_element="qm.multiple_inputs_for_element"
            in caps,
            supports_analog_delay="qm.analog_delay" in caps,
            supports_shared_oscillators="qm.shared_oscillators" in caps,
            supports_channel_weights="qm.channel_weights" in caps,
        )


def _has_job_streaming_state(
    qop_version: str, qua_implementation: Optional[QuaMachineInfo]
):
    if qop_version is not None:
        if qop_version.startswith("2.10"):
            return True
    if qua_implementation is not None:
        return "qm.job_streaming_state" in qua_implementation.capabilities
    return False



from qm.pb.inc_qua_pb2 import QuaProgram as _Q
from qm._loc import _get_loc


def var(name):
    """
    A reference to a variable
    :param name:
    :return:
    """
    exp = _Q.VarRefExpression()
    exp.name = name
    exp.loc = _get_loc()
    return exp


def binary(left, sop, right):
    """
    A binary operation
    :param left:
    :param sop:
    :param right:
    :return:
    """
    exp = _Q.AnyScalarExpression()
    exp.binaryOperation.SetInParent()
    exp.binaryOperation.loc = _get_loc()
    exp.binaryOperation.left.CopyFrom(left)
    exp.binaryOperation.right.CopyFrom(right)

    if sop == "+":
        op = _Q.BinaryExpression.ADD
    elif sop == "-":
        op = _Q.BinaryExpression.SUB
    elif sop == ">":
        op = _Q.BinaryExpression.GT
    elif sop == "<":
        op = _Q.BinaryExpression.LT
    elif sop == "<=":
        op = _Q.BinaryExpression.LET
    elif sop == ">=":
        op = _Q.BinaryExpression.GET
    elif sop == "==":
        op = _Q.BinaryExpression.EQ
    elif sop == "*":
        op = _Q.BinaryExpression.MULT
    elif sop == "/":
        op = _Q.BinaryExpression.DIV
    elif sop == "|":
        op = _Q.BinaryExpression.OR
    elif sop == "&":
        op = _Q.BinaryExpression.AND
    elif sop == "^":
        op = _Q.BinaryExpression.XOR
    elif sop == "<<":
        op = _Q.BinaryExpression.SHL
    elif sop == ">>":
        op = _Q.BinaryExpression.SHR
    else:
        raise Exception("Unsupported operator " + sop)

    exp.binaryOperation.op = op
    return exp


def literal_int(value):
    exp = _Q.AnyScalarExpression()
    exp.literal.value = str(value)
    exp.literal.type = _Q.INT
    exp.literal.loc = _get_loc()
    return exp


def literal_bool(value):
    exp = _Q.AnyScalarExpression()
    exp.literal.value = str(value)
    exp.literal.type = _Q.BOOL
    exp.literal.loc = _get_loc()
    return exp


def literal_real(value):
    exp = _Q.AnyScalarExpression()
    exp.literal.value = str(value)
    exp.literal.type = _Q.REAL
    exp.literal.loc = _get_loc()
    return exp


def io1():
    exp = _Q.AnyScalarExpression()
    exp.variable.ioNumber = 1
    exp.variable.loc = _get_loc()
    return exp


def io2():
    exp = _Q.AnyScalarExpression()
    exp.variable.ioNumber = 2
    exp.variable.loc = _get_loc()
    return exp


def array(value, index_exp):
    if index_exp is None:
        return value
    else:
        loc = _get_loc()
        exp = _Q.AnyScalarExpression()
        exp.arrayCell.arrayVar.CopyFrom(value)
        exp.arrayCell.arrayVar.loc = loc
        exp.arrayCell.index.CopyFrom(index_exp)
        exp.arrayCell.loc = loc
        return exp


def var_ref(value, index_exp):
    exp = _Q.AnyScalarExpression()

    loc = _get_loc()
    if index_exp is None:
        exp.variable.name = value
        exp.variable.loc = loc
    else:
        array_var = _Q.ArrayVarRefExpression()
        array_var.name = value
        exp.arrayCell.arrayVar.CopyFrom(array_var)
        exp.arrayCell.arrayVar.loc = loc
        exp.arrayCell.index.CopyFrom(index_exp)
        exp.arrayCell.loc = loc

    return exp


def lib_func(lib_name, func_name, *args):
    exp = _Q.AnyScalarExpression()

    exp.libFunction.SetInParent()
    exp.libFunction.loc = _get_loc()
    exp.libFunction.functionName = func_name
    exp.libFunction.libraryName = lib_name
    for arg in args:
        if isinstance(arg, _Q.ArrayVarRefExpression):
            element = exp.libFunction.arguments.add()
            element.array.CopyFrom(arg)
        else:
            element = exp.libFunction.arguments.add()
            element.scalar.CopyFrom(arg)

    return exp



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: qm/grpc/quantum_simulator/v1/physical_system.proto
# plugin: python-betterproto
from dataclasses import dataclass
from typing import Dict, List

import betterproto


@dataclass
class InputPort(betterproto.Message):
    """Input port of the quantum system (output port of OPX)"""

    # name of the controller
    controller: str = betterproto.string_field(1)
    # id
    id: int = betterproto.uint32_field(2)


@dataclass
class OutputPort(betterproto.Message):
    """Output port of the quantum system (input port of OPX)"""

    # name of the controller
    controller: str = betterproto.string_field(1)
    # id
    id: int = betterproto.uint32_field(2)


@dataclass
class TransmonProperties(betterproto.Message):
    """A class that holds the physical properties of a transmon."""

    # frequency of the qubit in Hz
    frequency: float = betterproto.double_field(1)
    # anharmonicity of the qubit in Hz
    anharmonicity: float = betterproto.double_field(2)
    # T1 time in ns
    t1: float = betterproto.double_field(3)
    # T2 time in ns
    t2: float = betterproto.double_field(4)
    n_levels: int = betterproto.uint32_field(5)
    n_order: int = betterproto.uint32_field(6)


@dataclass
class IQMixerIn(betterproto.Message):
    """IQMixer for up conversion"""

    # output port of the OPX
    i: "InputPort" = betterproto.message_field(1)
    # output port of the OPX
    q: "InputPort" = betterproto.message_field(2)
    # frequency of the local oscillator in Hz
    lo_frequency: float = betterproto.double_field(3)


@dataclass
class IQMixerOut(betterproto.Message):
    """IQMixer for down conversion"""

    # input port of the OPX
    i: "OutputPort" = betterproto.message_field(1)
    # input port of the OPX
    q: "OutputPort" = betterproto.message_field(2)
    # frequency of the local oscillator in Hz
    lo_frequency: float = betterproto.double_field(3)


@dataclass
class AnalogFrontend(betterproto.Message):
    """
    A class to describe analog frontend (currently just a constant
    multiplicative factor)
    """

    # a constant factor to account for losses or amplification of the signal
    amp_factor: float = betterproto.double_field(1)


@dataclass
class FluxLine(betterproto.Message):
    """
    A class to describe flux line and how it is connected to the OPX. The
    frequency of the qubit follows is tuned as: amp*(flux voltage - offset)^2
    """

    # output port of the OPX driving the flux line
    port: "InputPort" = betterproto.message_field(1)
    # amplitude in units of GHz/V^2
    amp_factor: float = betterproto.double_field(2)
    # offset voltage at which the bias is zero
    offset: float = betterproto.double_field(3)


@dataclass
class FixedFrequencyTransmon(betterproto.Message):
    pass


@dataclass
class TunableFrequencyTransmon(betterproto.Message):
    """A class to describe tunable frequency transmon"""

    # flux line controlling the transmon frequency
    flux_line: "FluxLine" = betterproto.message_field(1)


@dataclass
class Transmon(betterproto.Message):
    """
    Transmon contains the physical properties of the transmonand the
    information about how it is connected to the IQ Mixer and OPX
    """

    # physical properties of the transmon
    properties: "TransmonProperties" = betterproto.message_field(1)
    # IQ mixer used to drive the qubit
    i_q_mixer: "IQMixerIn" = betterproto.message_field(2)
    # analog frontend
    analog_frontend: "AnalogFrontend" = betterproto.message_field(3)
    # fixed frequency qubit
    fixed_frequency_transmon: "FixedFrequencyTransmon" = betterproto.message_field(
        4, group="type"
    )
    # tunable frequency qubit
    tunable_frequency_transmon: "TunableFrequencyTransmon" = betterproto.message_field(
        5, group="type"
    )


@dataclass
class DriveResonator(betterproto.Message):
    pass


@dataclass
class ReadoutResonator(betterproto.Message):
    """
    A class with the information about how the back reflected from the
    resonator is fed to the OPX.
    """

    i_q_mixer: "IQMixerOut" = betterproto.message_field(1)


@dataclass
class ResonatorProperties(betterproto.Message):
    # resonator frequency in Hz
    frequency: float = betterproto.double_field(1)
    # resonator loss rate in Hz
    lossrate: float = betterproto.double_field(2)
    # number of n_levels
    n_levels: int = betterproto.uint32_field(3)


@dataclass
class Resonator(betterproto.Message):
    """
    A class to describe the physical properties of readout resonator and how it
    is connected to the OPX
    """

    properties: "ResonatorProperties" = betterproto.message_field(1)
    # IQMixer to upconvert the drive
    i_q_mixer: "IQMixerIn" = betterproto.message_field(2)
    # analog frontend
    analog_frontend: "AnalogFrontend" = betterproto.message_field(3)
    # drive resonator
    drive_resonator: "DriveResonator" = betterproto.message_field(4, group="type")
    # readout resonator
    readout_resonator: "ReadoutResonator" = betterproto.message_field(5, group="type")


@dataclass
class CapacitiveCoupler(betterproto.Message):
    """A class for describing capacitive coupling between two qubits"""

    # output port of the OPX driving the coupler
    g: "InputPort" = betterproto.message_field(1)
    # amplitude factor in units of GHz/V
    amp_factor: float = betterproto.double_field(2)


@dataclass
class TransmonTransmonCoupling(betterproto.Message):
    """Transmon-transmon coupling"""

    # name of the capacitive coupler
    capacitive_coupler: str = betterproto.string_field(1)
    # name of the qubit 1
    qb1: str = betterproto.string_field(2)
    # name of the qubit 2
    qb2: str = betterproto.string_field(3)


@dataclass
class TransmonResonatorCoupling(betterproto.Message):
    """
    Transmon-resonator coupling describes the connectivity between a qubit and
    a resonator
    """

    # name of the qubit
    transmon: str = betterproto.string_field(1)
    # name of the resonator
    resonator: str = betterproto.string_field(2)
    # strength of the coupling in GHz
    g: float = betterproto.double_field(3)
    # coupling type
    is_dispersive: bool = betterproto.bool_field(4)


@dataclass
class SCChip(betterproto.Message):
    """
    A class to describe a superconducting chip layout consisting of fixed/flux
    tunable transmons, capacitive couplers and readout resonators.
    """

    # a dictionary of Transmons
    transmons: Dict[str, "Transmon"] = betterproto.map_field(
        1, betterproto.TYPE_STRING, betterproto.TYPE_MESSAGE
    )
    # a dictionary of resonators
    resonators: Dict[str, "Resonator"] = betterproto.map_field(
        2, betterproto.TYPE_STRING, betterproto.TYPE_MESSAGE
    )
    # a dictionary of capacitive couplers
    capacitive_couplers: Dict[str, "CapacitiveCoupler"] = betterproto.map_field(
        3, betterproto.TYPE_STRING, betterproto.TYPE_MESSAGE
    )
    # a list containing all the qubit-qubit couplings in the chip
    transmon_transmon_coupling: List[
        "TransmonTransmonCoupling"
    ] = betterproto.message_field(4)
    # a list containing the qubit-resonator couplings and their respective
    # strengths
    transmon_resonator_coupling: List[
        "TransmonResonatorCoupling"
    ] = betterproto.message_field(5)


@dataclass
class Offset(betterproto.Message):
    port: "InputPort" = betterproto.message_field(1)
    value: float = betterproto.double_field(2)


@dataclass
class Waveform(betterproto.Message):
    constant_waveform: "ConstantWaveform" = betterproto.message_field(1, group="type")
    arbitrary_waveform: "ArbitraryWaveform" = betterproto.message_field(2, group="type")


@dataclass
class ConstantWaveform(betterproto.Message):
    sample: float = betterproto.double_field(1)


@dataclass
class ArbitraryWaveform(betterproto.Message):
    samples: List[float] = betterproto.double_field(1)


@dataclass
class Error(betterproto.Message):
    type: str = betterproto.string_field(1)
    stack: str = betterproto.string_field(2)


@dataclass
class ColdAtomDevice(betterproto.Message):
    v: float = betterproto.double_field(1)
    qubit: str = betterproto.string_field(2)
    frequency_01: float = betterproto.double_field(3)
    frequency_1r: float = betterproto.double_field(4)
    positions: List["Coordinate"] = betterproto.message_field(5)
    t1: float = betterproto.double_field(6)
    t2: float = betterproto.double_field(7)


@dataclass
class Coordinate(betterproto.Message):
    x: float = betterproto.double_field(1)
    y: float = betterproto.double_field(2)


@dataclass
class ColdAtomSetup(betterproto.Message):
    device: "ColdAtomDevice" = betterproto.message_field(1)
    global_raman_beam: "GlobalRamanBeam" = betterproto.message_field(2)
    global_rydberg_beam: "GlobalRydbergBeam" = betterproto.message_field(3)
    local_raman_detuning: "ACStarkShifter" = betterproto.message_field(4)
    local_rydberg_detuning: "ACStarkShifter" = betterproto.message_field(5)


@dataclass
class BeamSteerer(betterproto.Message):
    a_o_d_x: str = betterproto.string_field(1)
    a_o_d_y: str = betterproto.string_field(2)
    reference_position: "Coordinate" = betterproto.message_field(3)
    eta_x: float = betterproto.double_field(4)
    eta_y: float = betterproto.double_field(5)


@dataclass
class ACStarkShifter(betterproto.Message):
    a_o_m: str = betterproto.string_field(1)
    delta: float = betterproto.double_field(2)
    eta_x: List[float] = betterproto.double_field(3)
    beam_steerer: "BeamSteerer" = betterproto.message_field(4)


@dataclass
class GlobalRamanBeam(betterproto.Message):
    intensity_modulator: str = betterproto.string_field(1)
    frequency: float = betterproto.double_field(2)
    scale_factor: float = betterproto.double_field(3)
    lo_frequency: float = betterproto.double_field(4)


@dataclass
class GlobalRydbergBeam(betterproto.Message):
    a_o_m1: str = betterproto.string_field(1)
    a_o_m2: str = betterproto.string_field(2)
    frequency1: float = betterproto.double_field(3)
    frequency2: float = betterproto.double_field(4)
    scale_factor1: float = betterproto.double_field(5)
    scale_factor2: float = betterproto.double_field(6)


@dataclass
class PhysicalConfig(betterproto.Message):
    chip: "SCChip" = betterproto.message_field(1, group="type")
    cold_atom_setup: "ColdAtomSetup" = betterproto.message_field(2, group="type")



import dataclasses
from typing import Optional, List, Tuple

import grpc
import grpclib.client
from google.protobuf import empty_pb2
import grpclib.events
from qua import QuaClient
from qua.info import QuaMachineInfo

from qm.run_async import run_async
from qm.capabilities import ServerCapabilities
from qm.grpc_client_interceptor import header_manipulator_client_interceptor
from qm.pb import frontend_pb2_grpc


@dataclasses.dataclass
class DetectedServer:
    port: int
    host: str
    qop_version: Optional[str]
    grpc_channel: Optional[grpc.Channel]
    # does it implement the QUA service
    qua_implementation: Optional[QuaMachineInfo]
    qua_client: QuaClient

    def close_channels(self):
        if self.grpc_channel is not None:
            self.grpc_channel.close()

    @property
    def capabilities(self) -> ServerCapabilities:
        return ServerCapabilities.build(
            qop_version=self.qop_version, qua_implementation=self.qua_implementation
        )


def default_interceptors(user_token):
    return [
        header_manipulator_client_interceptor.header_adder_interceptor(
            "x-grpc-service", "gateway"
        ),
        header_manipulator_client_interceptor.header_adder_interceptor(
            "authorization", f"Bearer {user_token}"
        ),
    ]


def detect_server(
    user_token,
    credentials,
    host,
    port_from_user_config: int,
    user_provided_port: Optional[int],
) -> Tuple[Optional[DetectedServer], List[int]]:
    max_message_size = 1024 * 1024 * 100  # 100 mb in bytes
    options = [
        ("grpc.max_receive_message_length", max_message_size),
    ]
    detected = None
    ports_to_try = (
        [user_provided_port]
        if user_provided_port is not None
        else [port_from_user_config, 80]
    )
    for port in ports_to_try:
        detected = try_server(user_token, credentials, options, host, port)
        if detected.qop_version is not None:
            break

    if detected.qop_version is None:
        return None, ports_to_try
    else:
        return detected, ports_to_try


def try_server(user_token, credentials, options, host, port) -> DetectedServer:
    interceptors = default_interceptors(user_token)
    intercept_channel = build_channel(interceptors, credentials, options, host, port)
    frontend = frontend_pb2_grpc.FrontendStub(intercept_channel)

    detected = DetectedServer(
        port=port,
        host=host,
        qop_version=None,
        qua_implementation=None,
        grpc_channel=intercept_channel,
        qua_client=QuaClient(
            host=host,
            port=port,
            grpc_channel_factory=lambda: build_grpclib_channel(
                credentials, user_token, host, port
            ),
        ),
    )
    try:
        req = empty_pb2.Empty()
        res = frontend.GetVersion(req)
        detected.qop_version = str(res.value)
    except grpc.RpcError:
        pass
    try:
        info: QuaMachineInfo = run_async(detected.qua_client.get_server_info())
        detected.qua_implementation = info
    except ConnectionRefusedError:
        pass
    return detected


def build_channel(interceptors, credentials, options, host, port):
    address = host + ":" + str(port)
    compression = grpc.Compression.Gzip
    channel = (
        grpc.insecure_channel(
            target=address,
            options=options,
            compression=compression,
        )
        if credentials is None
        else grpc.secure_channel(
            target=address,
            credentials=credentials,
            options=options,
            compression=compression,
        )
    )
    return grpc.intercept_channel(channel, *interceptors)


def build_grpclib_channel(credentials, user_token, host, port):
    channel = grpclib.client.Channel(
        host=host,
        port=port,
        ssl=credentials is not None,
        config=grpclib.client.Configuration(),
    )

    async def send_request(event: grpclib.events.SendRequest):
        event.metadata["x-grpc-service"] = "gateway"
        event.metadata["authorization"] = f"Bearer {user_token}"

    grpclib.events.listen(channel, grpclib.events.SendRequest, send_request)
    return channel



import time
import zipfile
from io import BufferedWriter, BytesIO
from typing import Optional, Union, Dict, Tuple, Generator, List
import numpy
import numpy.lib.format as _format
import json as _json
from dataclasses import dataclass

from qm.StreamMetadata import (
    StreamMetadata,
    StreamMetadataError,
    _get_stream_metadata_dict_from_proto_resp,
)
from qm._errors import InvalidStreamMetadataError
from qm.logger import logger
from qm.pb.job_results_pb2_grpc import JobResultsServiceStub
from qm.pb.job_results_pb2 import (
    GetJobNamedResultHeaderRequest,
    GetJobNamedResultRequest,
    GetJobResultSchemaRequest,
    GetJobResultSchemaResponse,
    GetJobDebugDataRequest,
    GetProgramMetadataRequest,
    GetJobStateRequest,
)
from qm.persistence import BaseStore

TIMESTAMPS_LEGACY_EXT = "_timestamps"


def _parse_dtype(simple_dtype: str) -> dict:
    def hinted_tuple_hook(obj):
        if "__tuple__" in obj:
            return tuple(obj["items"])
        else:
            return obj

    dtype = _json.loads(simple_dtype, object_hook=hinted_tuple_hook)
    return dtype


@dataclass
class JobResultItemSchema:
    name: str
    dtype: dict
    shape: Tuple[int]
    is_single: bool
    expected_count: int


@dataclass
class JobResultSchema:
    items: Dict[str, JobResultItemSchema]


@dataclass
class NamedJobResultHeader:
    count_so_far: int
    is_single: bool
    output_name: str
    job_id: str
    d_type: dict
    shape: Tuple[int]
    has_dataloss: bool


@dataclass
class JobStreamingState:
    job_id: str
    done: bool
    closed: bool
    has_dataloss: bool


class BaseNamedJobResult:
    def __init__(
        self,
        job_id: str,
        schema: JobResultItemSchema,
        service: JobResultsServiceStub,
        store: BaseStore,
        stream_metadata_errors: List[StreamMetadataError],
        stream_metadata: StreamMetadata,
    ) -> None:
        super().__init__()
        self._job_id = job_id
        self._schema = schema
        self._service = service
        self._store = store
        self._stream_metadata_errors = stream_metadata_errors
        self._stream_metadata = stream_metadata

    @property
    def name(self) -> str:
        """The name of result this handle is connected to"""
        return self._schema.name

    @property
    def job_id(self) -> str:
        """The job id this result came from"""
        return self._job_id

    @property
    def expected_count(self) -> int:
        return self._schema.expected_count

    @property
    def numpy_dtype(self):
        return self._schema.dtype

    @property
    def stream_metadata(self) -> StreamMetadata:
        """Provides the StreamMetadata of this stream.

        Metadata currently includes the values and shapes of the automatically identified loops
        in the program.

        """
        if len(self._stream_metadata_errors) > 0:
            logger.error("Error creating stream metadata:")
            for x in self._stream_metadata_errors:
                logger.error(f"{x.error} in {x.location}")
            raise InvalidStreamMetadataError(self._stream_metadata_errors)
        return self._stream_metadata

    def save_to_store(
        self, writer: Optional[Union[BufferedWriter, BytesIO, str]] = None, **kwargs
    ) -> int:
        """Saving to persistent store the NPY data of this result handle

        :param writer: An optional writer to override the store defined in \
            :class:`QuantumMachinesManager<qm.QuantumMachinesManager.QuantumMachinesManager>`

        :key flat_struct: results will have a flat structure - dimensions will be \
            part of the shape and not of the type

        :return: The number of items saved

        """
        kwargs.get("flat_struct")  # for hints
        own_writer = False
        if writer is None:
            own_writer = True
            writer = self._store.job_named_result(
                self._job_id, self._schema.name
            ).for_writing()
        try:
            header = self._get_named_header(**kwargs)
            return self._save_to_file(header, writer)
        finally:
            if own_writer:
                writer.close()

    def wait_for_values(self, count: int = 1, timeout: Optional[float] = None):
        """Wait until we know at least `count` values were processed for this named result

        :param count: The number of items to wait for
        :param timeout: Timeout for waiting in seconds
        :return:
        """
        start = time.time()
        end = start + max(0.0, timeout) if timeout is not None else None
        while end is None or time.time() < end:
            if self.count_so_far() >= count:
                return
            time.sleep(0.01)

    def wait_for_all_values(self, timeout: Optional[float] = None) -> bool:
        """Wait until we know all values were processed for this named result

        :param timeout: Timeout for waiting in seconds
        :return: True if job finished successfully and False if job has closed before done
        """
        start = time.time()
        end = start + max(0.0, timeout) if timeout is not None else None
        while end is None or time.time() < end:
            header = self._get_job_state()
            if header.done or header.closed:
                return header.done
            time.sleep(0.01)
        raise TimeoutError(f"result {self.name} was not done in time")

    def is_processing(self) -> bool:
        header = self._get_job_state()
        return not (header.done or header.closed)

    def count_so_far(self) -> int:
        """
        also `len(handle)`

        :return: The number of values this result has so far
        """
        return self._get_named_header().count_so_far

    def __len__(self) -> int:
        return self.count_so_far()

    def has_dataloss(self) -> bool:
        """
        :return: if there was data loss during job execution
        """
        return self._get_job_state().has_dataloss

    def _write_header(
        self,
        writer: Union[BufferedWriter, BytesIO, str],
        shape: Tuple[int],
        d_type: object,
    ):
        _format.write_array_header_2_0(
            writer, {"descr": d_type, "fortran_order": False, "shape": shape}
        )

    def _save_to_file(
        self, header: NamedJobResultHeader, writer: Union[BufferedWriter, BytesIO, str]
    ) -> int:
        count = 0
        request = GetJobNamedResultRequest()
        request.jobId = self._job_id
        request.outputName = self.name
        request.longOffset.value = 0
        request.limit = header.count_so_far

        owning_writer = False
        if type(writer) is str:
            writer = open(writer, "wb+")
            owning_writer = True

        try:
            finalShape = self._get_final_shape(header.count_so_far, header.shape)

            self._write_header(writer, finalShape, header.d_type)
            for result in self._service.GetJobNamedResult(request):
                count += result.countOfItems
                writer.write(result.data)

        finally:
            if owning_writer:
                writer.close()
        return count

    def _get_job_state(self, **kwargs) -> JobStreamingState:
        request = GetJobStateRequest()
        request.jobId = self._job_id
        response = self._service.GetJobState(request)
        return JobStreamingState(
            job_id=self._job_id,
            done=response.done,
            closed=response.closed,
            has_dataloss=response.hasDataloss,
        )

    def _get_named_header(self, **kwargs) -> NamedJobResultHeader:
        request = GetJobNamedResultHeaderRequest()
        request.jobId = self._job_id
        request.outputName = self.name
        if kwargs.get("flat_struct", False) is True:
            request.flatFormat = True
        response = self._service.GetJobNamedResultHeader(request)
        dtype = _parse_dtype(response.simpleDType)

        if (
            kwargs.get("check_for_errors", False)
            and response.HasField("hasExecutionErrors")
            and response.hasExecutionErrors.value
        ):
            logger.error(
                "Errors were detected, please fetch the execution report for more "
                "information."
            )

        return NamedJobResultHeader(
            count_so_far=response.countSoFar,
            is_single=response.isSingle,
            output_name=self.name,
            job_id=self.job_id,
            d_type=dtype,
            shape=tuple(response.shape),
            has_dataloss=response.hasDataloss,
        )

    def fetch_all(self, **kwargs):
        """
        Fetch a result from the current result stream saved in server memory.
        The result stream is populated by the save() and save_all() statements.
        Note that if save_all() statements are used, calling this function twice
        may give different results.

        :key flat_struct: results will have a flat structure - dimensions will be \
            part of the shape and not of the type

        :return: all result of current result stream

        """
        kwargs.get("flat_struct")  # for hints
        return self.fetch(slice(0, len(self)), **kwargs)

    def fetch(self, item: Union[int, slice], **kwargs) -> numpy.array:
        """
        Fetch a result from the current result stream saved in server memory.
        The result stream is populated by the save() and save_all() statements.
        Note that if save_all() statements are used, calling this function twice
        with the same item index may give different results.

        :param item: The index of the result in the saved results stream.

        :key flat_struct: results will have a flat structure - dimensions will be \
            part of the shape and not of the type

        :return: a single result if item is integer or multiple results if item is Python slice object.

        Example::
        >>>res.fetch(0)         #return the item in the top position
        >>>res.fetch(1)         #return the item in position number 2
        >>>res.fetch(slice(1,6))# return items from position 1 to position 6 (exclusive)
        >>>                     # same as res.fetch_all()[1:6]

        """
        kwargs.get("flat_struct")  # for hints
        if type(item) is int:
            start = item
            stop = item + 1
            step = None
        elif type(item) is slice:
            start = item.start
            stop = item.stop
            step = item.step
        else:
            raise Exception("fetch supports only int or slice")

        if step != 1 and step is not None:
            raise Exception("fetch supports step=1 or None in slices")

        if kwargs.get("check_for_errors") is None:
            header = self._get_named_header(**kwargs, check_for_errors=True)
        else:
            header = self._get_named_header(**kwargs)

        if stop is None:
            stop = header.count_so_far
        if start is None:
            start = 0

        writer = BytesIO()
        count = 0
        request = GetJobNamedResultRequest()
        request.jobId = self._job_id
        request.outputName = self._schema.name
        request.longOffset.value = start
        request.limit = stop - start

        data_writer = BytesIO()
        for result in self._service.GetJobNamedResult(request):
            count += result.countOfItems
            data_writer.write(result.data)

        finalShape = self._get_final_shape(count, header.shape)

        self._write_header(writer, finalShape, header.d_type)

        data_writer.seek(0)
        for d in data_writer:
            writer.write(d)

        writer.seek(0)

        if header.has_dataloss:
            logger.warning(f"Results variable {self.name} has data loss")

        return numpy.load(writer)

    @staticmethod
    def _get_final_shape(count, shape):
        if count == 1:
            finalShape = shape
        else:
            if len(shape) == 1 and shape[0] == 1:
                finalShape = (count,)
            else:
                finalShape = (count,) + shape
        return finalShape


class MultipleNamedJobResult(BaseNamedJobResult):
    """
    A handle to a result of a pipeline terminating with ``save_all``
    """

    def __init__(
        self,
        job_id: str,
        schema: JobResultItemSchema,
        service: JobResultsServiceStub,
        store: BaseStore,
        stream_metadata_errors: List[StreamMetadataError],
        jobResults,
        stream_metadata: StreamMetadata,
    ) -> None:
        if schema.is_single:
            raise Exception("expecting a multi-result schema")
        self.jobResults = jobResults
        super().__init__(
            job_id, schema, service, store, stream_metadata_errors, stream_metadata
        )

    def _wait_for_any_value(self):
        self.wait_for_values(count=1, timeout=0.5)

    def fetch(self, item: Union[int, slice], **kwargs) -> numpy.array:
        """
        Fetch a result from the current result stream saved in server memory.
        The result stream is populated by the save() and save_all() statements.
        Note that if save_all() statements are used, calling this function twice
        with the same item index may give different results.

        :param item: The index of the result in the saved results stream.

        :key flat_struct: results will have a flat structure - dimensions will be \
            part of the shape and not of the type

        :return: a single result if item is integer or multiple results if item is Python slice object.

        Example::
        >>>res.fetch(0)         #return the item in the top position
        >>>res.fetch(1)         #return the item in position number 2
        >>>res.fetch(slice(1,6))# return items from position 1 to position 6 (exclusive)
        >>>                     # same as res.fetch_all()[1:6]

        """
        if kwargs.get("flat_struct"):
            return super().fetch(item, **kwargs)
        else:
            # legacy support - reconstruct the old structure
            name = self._schema.name
            timestampsName = name + TIMESTAMPS_LEGACY_EXT
            timestampsResultHandle = self.jobResults.get(timestampsName)
            if timestampsResultHandle is None:
                return super().fetch(item, **kwargs)
            else:
                valuesResult = super().fetch(item, **kwargs, flat_struct=True)

                fetchedLength = len(valuesResult)
                if isinstance(item, slice):
                    item = slice(item.start, item.start + fetchedLength, item.step)
                else:
                    item = slice(0, fetchedLength, 1)

                timestampsResult = timestampsResultHandle.fetch(
                    item, **kwargs, flat_struct=True, check_for_errors=False
                )

                dtype = [
                    ("value", valuesResult.dtype),
                    ("timestamp", timestampsResult.dtype),
                ]  # timestampsResult.dtype.descr
                combined = numpy.rec.fromarrays(
                    [valuesResult, timestampsResult], dtype=dtype
                )
                return combined.view(numpy.ndarray).astype(dtype)

    def save_to_store(
        self, writer: Optional[Union[BufferedWriter, BytesIO, str]] = None, **kwargs
    ) -> int:
        """Saving to persistent store the NPY data of this result handle

        :param writer: An optional writer to override the store defined in \
                        :class:`QuantumMachinesManager<qm.QuantumMachinesManager.QuantumMachinesManager>`

        :return: The number of items saved

        """
        if kwargs.get("flat_struct"):
            return super().save_to_store(writer, **kwargs)
        else:
            # legacy support - reconstruct the old structure
            name = self._schema.name
            timestampsName = name + TIMESTAMPS_LEGACY_EXT
            timestampsResultHandle = self.jobResults.get(timestampsName)
            if timestampsResultHandle is None:
                return super().save_to_store(writer, **kwargs)
            else:
                final_result = self.fetch_all(**kwargs)
                own_writer = False
                if writer is None:
                    own_writer = True
                    writer = self._store.job_named_result(
                        self._job_id, self._schema.name
                    ).for_writing()
                try:
                    owning_writer = False
                    if type(writer) is str:
                        writer = open(writer, "wb+")
                        owning_writer = True

                    try:
                        self._write_header(
                            writer, (len(final_result),), final_result.dtype.descr
                        )
                        writer.write(final_result.tobytes())

                    finally:
                        if owning_writer:
                            writer.close()
                    return len(final_result)
                finally:
                    if own_writer:
                        writer.close()


class SingleNamedJobResult(BaseNamedJobResult):
    """
    A handle to a result of a pipeline terminating with ``save``
    """

    def __init__(
        self,
        job_id: str,
        schema: JobResultItemSchema,
        service: JobResultsServiceStub,
        store: BaseStore,
        stream_metadata_errors: List[StreamMetadataError],
        stream_metadata: StreamMetadata,
    ) -> None:
        if not schema.is_single:
            raise Exception("expecting a single-result schema")
        super().__init__(
            job_id, schema, service, store, stream_metadata_errors, stream_metadata
        )

    def wait_for_values(self, count: int = 1, timeout: Optional[float] = None):
        if count != 1:
            raise RuntimeError("single result can wait only for a single value")
        return super().wait_for_values(1, timeout)

    def fetch_all(self, **kwargs):
        """
        Fetch a result from the current result stream saved in server memory.
        The result stream is populated by the save() and save_all() statements.
        Note that if save_all() statements are used, calling this function twice
        may give different results.

        :key flat_struct: results will have a flat structure - dimensions will be \
            part of the shape and not of the type

        :return: all result of current result stream

        """
        kwargs.get("flat_struct")  # for hints
        return self.fetch(0, **kwargs)

    def fetch(self, item: Union[int, slice], **kwargs):

        """
        Fetch a single result from the current result stream saved in server memory.
        The result stream is populated by the save().

        :param item: ignored

        :key flat_struct: results will have a flat structure - dimensions will be \
            part of the shape and not of the type

        :return: the current result

        Example::
        >>>res.fetch() #return the item in the top position

        """
        if (isinstance(item, int) and item != 0) or isinstance(item, slice):
            logger.warning("Fetching single result will always return the single value")
        value = super().fetch(0, **kwargs)
        if kwargs.get("flat_struct", False) is True:
            if len(value) == 0:
                return None
            elif len(value) == 1:
                return value[0]
            else:
                return value
        else:
            if len(value) == 0:
                return None
            elif len(value[0]) == 1:
                return value[0][0]
            else:
                return value[0]


class JobResults:
    """
    Access to the results of a QmJob

    This object is created by calling :attr:`QmJob.result_handles<qm.QmJob.QmJob.result_handles>`

    Assuming you have an instance of JobResults::

        job_results:JobResults

    This object is iterable::

        for name, handle in job_results:
            print(name)

    Can detect if a name exists::

        if "somename" in job_results:
            print("somename exists!")
            handle = job_results.get("somename")

    """

    def __init__(
        self, job_id: str, service: JobResultsServiceStub, store: BaseStore
    ) -> None:
        super().__init__()
        self._job_id = job_id
        self._service = service
        self._store = store
        schema = JobResults._load_schema(job_id, service)
        self._schema = schema
        (stream_metadata_errors, stream_metadata_dict) = self._get_stream_metadata()

        self._all_results: Dict[str, BaseNamedJobResult] = {}
        for (name, item_schema) in schema.items.items():
            stream_metadata = stream_metadata_dict.get(name)
            if item_schema.is_single:
                result = SingleNamedJobResult(
                    job_id,
                    item_schema,
                    service,
                    store,
                    stream_metadata_errors,
                    stream_metadata,
                )
            else:
                result = MultipleNamedJobResult(
                    job_id,
                    item_schema,
                    service,
                    store,
                    stream_metadata_errors,
                    self,
                    stream_metadata,
                )
            self._all_results[name] = result
            if not hasattr(self, name):
                setattr(self, name, result)

    def _get_stream_metadata(
        self,
    ) -> Optional[Tuple[List[StreamMetadataError], Dict[str, StreamMetadata]]]:
        request = GetProgramMetadataRequest()
        request.jobId = self._job_id
        program_metadata = self._service.GetProgramMetadata(request)

        if program_metadata.success:
            stream_metadata_errors = [
                StreamMetadataError(x.error, x.location)
                for x in program_metadata.programStreamMetadata.streamMetadataExtractionError
            ]

            stream_metadata_dict = {}
            if len(stream_metadata_errors) == 0:
                stream_metadata_dict = _get_stream_metadata_dict_from_proto_resp(
                    program_metadata
                )

            return stream_metadata_errors, stream_metadata_dict

        return None

    def __iter__(self) -> Generator[Tuple[str, BaseNamedJobResult], None, None]:
        for item in self._schema.items.values():
            yield item.name, self.get(item.name)

    def is_processing(self) -> bool:
        """Check if the job is still processing results

        :return: True if results are still being processed, False otherwise
        """
        key = list(self._all_results.keys())[0]
        return self._all_results[key].is_processing()

    def save_to_store(
        self, writer: Optional[Union[BufferedWriter, BytesIO, str]] = None, **kwargs
    ):
        """Save all results to store (file system by default) in a single NPZ file

        :param writer: An optional writer to be used instead of the pre-populated \
            store passed to :class:`qm.QuantumMachinesManager.QuantumMachinesManager`

        :key flat_struct: results will have a flat structure - dimensions will be \
            part of the shape and not of the type

        """
        kwargs.get("flat_struct")  # for hints
        own_writer = False
        if writer is None:
            own_writer = True
            writer = self._store.all_job_results(self._job_id).for_writing()
        zipf = None
        try:
            zipf = zipfile.ZipFile(
                writer, allowZip64=True, mode="w", compression=zipfile.ZIP_DEFLATED
            )
            for (name, result) in self:
                with zipf.open(f"{name}.npy", "w") as entry:
                    result.save_to_store(entry, **kwargs)
                pass
        finally:
            zipf.close()
            if own_writer:
                writer.close()

    @staticmethod
    def _load_schema(job_id: str, service: JobResultsServiceStub) -> JobResultSchema:
        request = GetJobResultSchemaRequest()
        request.jobId = job_id
        response: GetJobResultSchemaResponse = service.GetJobResultSchema(request)
        return JobResultSchema(
            {
                item.name: JobResultItemSchema(
                    item.name,
                    _parse_dtype(item.simpleDType),
                    tuple(item.shape),
                    item.isSingle,
                    item.expectedCount,
                )
                for item in response.items
            }
        )

    def get(
        self, name: str
    ) -> Optional[Union[MultipleNamedJobResult, SingleNamedJobResult]]:
        """Get a handle to a named result from :func:`stream_processing<~qm.qua._dsl.stream_processing>`

        :param name: The named result using in :func:`stream_processing<~qm.qua._dsl.stream_processing>`
        :return: A handle object to the results :class:`MultipleNamedJobResult` or :class:`SingleNamedJobResult` \
                or None if the named results in unknown

        """
        return self._all_results[name] if name in self._all_results else None

    def __contains__(self, name: str):
        return name in self._all_results

    def wait_for_all_values(self, timeout: Optional[float] = None) -> bool:
        """Wait until we know all values were processed for all named results

        :param timeout: Timeout for waiting in seconds
        :return: True if all finished successfully, False if any result was closed before done
        """
        start = time.time()
        end: Optional[int] = start + max(0.0, timeout) if timeout is not None else None
        while end is None or time.time() < end:
            header = self._get_job_state()
            if header.done or header.closed:
                return header.done
            time.sleep(0.01)
        raise TimeoutError("job was not done in time")

    def _get_job_state(self) -> JobStreamingState:
        request = GetJobStateRequest()
        request.jobId = self._job_id
        response = self._service.GetJobState(request)
        return JobStreamingState(
            job_id=self._job_id,
            done=response.done,
            closed=response.closed,
            has_dataloss=response.hasDataloss,
        )

    def get_debug_data(
        self, writer: Optional[Union[BufferedWriter, BytesIO, str]] = None
    ):
        """

        :return: debugging data to report to QM
        """
        request = GetJobDebugDataRequest()
        request.jobId = self._job_id

        if writer is None:
            writer = f"./{self._job_id}-DebugData.zip"

        owning_writer = False
        if type(writer) is str:
            writer = open(writer, "wb+")
            owning_writer = True

        try:
            for result in self._service.GetJobDebugData(request):
                writer.write(result.data)

        finally:
            if owning_writer:
                writer.close()



from dataclasses import dataclass
from typing import Union, TypeVar, Generic, NewType, List

from qm.pb.frontend_pb2 import SimulationRequest


class SimulationConfig(object):
    """
    Creates a configuration object to pass to
    :func:`~qm.QuantumMachinesManager.QuantumMachinesManager.simulate`

    :param int duration: The duration to run the simulation for, in clock cycles
    :param bool include_analog_waveforms: True to collect simulated analog waveform names
    :param bool include_digital_waveforms: True to collect simulated digital waveform names
    :param SimulatorInterface simulation_interface:

        Interface for to simulator. Currently supported interfaces:

            ``None`` - Zero inputs

            :class:`~qm.simulate.loopback.LoopbackInterface` - Loopback output to input

            :class:`~qm.simulate.raw.RawInterface` - Specify samples for inputs

    :param  List[ControllerConnection] controller_connections: A list of connections
        between the controllers in the config
    :param int extraProcessingTimeoutInMs: timeout in ms to wait for stream processing
        to finish. Default is -1, which is disables the timeout
    """

    duration = 0
    simulate_analog_outputs = False

    def __init__(
        self,
        duration=0,
        include_analog_waveforms=False,
        include_digital_waveforms=False,
        simulation_interface=None,
        controller_connections: List["ControllerConnection"] = None,
        extraProcessingTimeoutInMs=-1,
    ):
        if controller_connections is None:
            controller_connections = []
        super(SimulationConfig, self).__init__()
        self.duration = duration
        self.include_analog_waveforms = include_analog_waveforms is True
        self.include_digital_waveforms = include_digital_waveforms is True
        self.simulation_interface = simulation_interface
        self.controller_connections = controller_connections
        self.extraProcessingTimeoutInMs = extraProcessingTimeoutInMs

    def update_simulate_request(self, request: SimulationRequest):
        if self.simulation_interface is None:
            request.simulate.simulationInterface.none.SetInParent()
        else:
            self.simulation_interface.update_simulate_request(request)


class SimulatorInterface(object):
    def update_simulate_request(self, request: SimulationRequest):
        pass


@dataclass
class InterOpxAddress:
    """
    :param str controller: The name of the controller
    :param bool is_left_connection:
    """

    controller: str
    is_left_connection: bool


@dataclass
class InterOpxChannel:
    """
    :param str controller: The name of the controller
    :param int channel_number:
    """

    controller: str
    channel_number: int


InterOpxPairing = NewType("InterOpxPairing", Union[InterOpxAddress, InterOpxChannel])
T = TypeVar("T", InterOpxPairing, InterOpxPairing)


class ControllerConnection(Generic[T]):
    """"""

    def __init__(self, source: T, target: T):
        self.source = source
        self.target = target



import time
from datetime import datetime
from typing import Optional

from qm.QmJob import QmJob
from qm._errors import JobCancelledError
from qm.pb.frontend_pb2 import GetJobExecutionStatusRequest


class QmPendingJob(object):
    """
    A Class describing a job in the execution queue
    """

    def __init__(
        self,
        job_id,
        machine: "qm.QuantumMachine.QuantumMachine",
        qmm: "qm.QuantumMachinesManager.QuantumMachinesManager",
    ):

        self._id = job_id
        self._machine = machine
        self._qmm = qmm
        self._added_user_id = None
        self._time_added = None
        self._set_metadata()

    def _set_metadata(self):
        response = self._get_execution_status()
        if response.status.HasField("pending"):
            self._added_user_id = response.status.pending.addedBy
            self._time_added = datetime.fromtimestamp(
                response.status.pending.timeAdded.seconds
            )
        elif response.status.HasField("running"):
            self._added_user_id = response.status.running.addedBy
            self._time_added = datetime.fromtimestamp(
                response.status.running.timeAdded.seconds
            )
        elif response.status.HasField("completed"):
            self._added_user_id = response.status.completed.addedBy
            self._time_added = datetime.fromtimestamp(
                response.status.completed.timeAdded.seconds
            )
        elif response.status.HasField("loading"):
            self._added_user_id = response.status.loading.addedBy
            self._time_added = datetime.fromtimestamp(
                response.status.loading.timeAdded.seconds
            )

    def id(self) -> str:
        """
        :return: The id of the job

        """
        return self._id

    @property
    def user_added(self):
        return self._added_user_id

    @property
    def time_added(self) -> datetime:
        return self._time_added

    def position_in_queue(self) -> Optional[int]:
        """
        Returns the current position of the job in the queue

        :return: The position in the queue

        Example:

            >>> pending_job.position_in_queue()

        """
        response = self._get_execution_status()
        if response.status.HasField("pending"):
            return response.status.pending.positionInQueue
        return None  # Job left queue

    def _get_execution_status(self):
        request = GetJobExecutionStatusRequest()
        request.quantumMachineId = self._machine.id
        request.jobId = self._id
        response = self._qmm._frontend.GetJobExecutionStatus(request)
        return response

    def wait_for_execution(self, timeout: Optional[float] = None) -> QmJob:
        """
        Waits for a job to complete or until the timeout has elapsed

        :param timeout: Timeout (in seconds) for this operation

        :return: The completed ``QmJob``

        """
        start = time.time()
        end = start + max(0.0, timeout) if timeout is not None else None

        while end is None or time.time() < end:
            response = self._get_execution_status()
            if response.status.HasField("pending") or response.status.HasField(
                "loading"
            ):
                time.sleep(0.02)
            elif response.status.HasField("running") or response.status.HasField(
                "completed"
            ):
                return QmJob(self._qmm, self._id, None)
            else:
                raise JobCancelledError(f"job {self._id} was cancelled")
        raise TimeoutError(f"job {self._id} was not started in time")

    def cancel(self) -> bool:
        """
        Removes the job from the queue

        :return: true if the operation was successful

        Example:

            >>> pending_job.cancel()

        """
        return self._machine.queue.remove_by_id(self._id) > 0



from collections import Iterable

from google.protobuf.struct_pb2 import NULL_VALUE, Value


def python_to_pb(data):
    out = Value()
    if data is None:
        out.null_value = NULL_VALUE
    elif type(data) is str:
        out.string_value = data
    elif type(data) is int:
        out.number_value = data
    elif type(data) is float:
        out.number_value = data
    elif type(data) is dict:
        out.struct_value.SetInParent()
        for k, v in data.items():
            out.struct_value.fields.get_or_create(str(k)).CopyFrom(python_to_pb(v))
    elif isinstance(data, Iterable):
        for item in data:
            out.list_value.values.add().CopyFrom(python_to_pb(item))
    else:
        raise Exception("Unhandled type: " + str(type(data)))
    return out



from qm.QuantumMachine import QuantumMachine  # noqa
from qm.program import _Program  # noqa
from qm.program import _ResultAnalysis  # noqa
from qm.program._qua_config_schema import validate_config  # noqa
from .generate_qua_script import generate_qua_script  # noqa
from qm.simulate import *  # noqa



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: qm/pb/general_messages.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='qm/pb/general_messages.proto',
  package='qm.grpc.general_messages',
  syntax='proto3',
  serialized_options=b'\n\027qm.grpc.generalMessagesB\024GeneralMessagesProtoP\001\242\002\003HLW',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x1cqm/pb/general_messages.proto\x12\x18qm.grpc.general_messages\"-\n\x0c\x45rrorMessage\x12\x0c\n\x04\x63ode\x18\x01 \x01(\x05\x12\x0f\n\x07message\x18\x02 \x01(\t\"<\n\x06Matrix\x12\x0b\n\x03v00\x18\x01 \x01(\x01\x12\x0b\n\x03v01\x18\x02 \x01(\x01\x12\x0b\n\x03v10\x18\x03 \x01(\x01\x12\x0b\n\x03v11\x18\x04 \x01(\x01*Z\n\x0cMessageLevel\x12\x17\n\x13Message_LEVEL_ERROR\x10\x00\x12\x19\n\x15Message_LEVEL_WARNING\x10\x01\x12\x16\n\x12Message_LEVEL_INFO\x10\x02\x42\x37\n\x17qm.grpc.generalMessagesB\x14GeneralMessagesProtoP\x01\xa2\x02\x03HLWb\x06proto3'
)

_MESSAGELEVEL = _descriptor.EnumDescriptor(
  name='MessageLevel',
  full_name='qm.grpc.general_messages.MessageLevel',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='Message_LEVEL_ERROR', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='Message_LEVEL_WARNING', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='Message_LEVEL_INFO', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=167,
  serialized_end=257,
)
_sym_db.RegisterEnumDescriptor(_MESSAGELEVEL)

MessageLevel = enum_type_wrapper.EnumTypeWrapper(_MESSAGELEVEL)
Message_LEVEL_ERROR = 0
Message_LEVEL_WARNING = 1
Message_LEVEL_INFO = 2



_ERRORMESSAGE = _descriptor.Descriptor(
  name='ErrorMessage',
  full_name='qm.grpc.general_messages.ErrorMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='code', full_name='qm.grpc.general_messages.ErrorMessage.code', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='message', full_name='qm.grpc.general_messages.ErrorMessage.message', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=58,
  serialized_end=103,
)


_MATRIX = _descriptor.Descriptor(
  name='Matrix',
  full_name='qm.grpc.general_messages.Matrix',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='v00', full_name='qm.grpc.general_messages.Matrix.v00', index=0,
      number=1, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='v01', full_name='qm.grpc.general_messages.Matrix.v01', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='v10', full_name='qm.grpc.general_messages.Matrix.v10', index=2,
      number=3, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='v11', full_name='qm.grpc.general_messages.Matrix.v11', index=3,
      number=4, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=105,
  serialized_end=165,
)

DESCRIPTOR.message_types_by_name['ErrorMessage'] = _ERRORMESSAGE
DESCRIPTOR.message_types_by_name['Matrix'] = _MATRIX
DESCRIPTOR.enum_types_by_name['MessageLevel'] = _MESSAGELEVEL
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ErrorMessage = _reflection.GeneratedProtocolMessageType('ErrorMessage', (_message.Message,), {
  'DESCRIPTOR' : _ERRORMESSAGE,
  '__module__' : 'qm.pb.general_messages_pb2'
  # @@protoc_insertion_point(class_scope:qm.grpc.general_messages.ErrorMessage)
  })
_sym_db.RegisterMessage(ErrorMessage)

Matrix = _reflection.GeneratedProtocolMessageType('Matrix', (_message.Message,), {
  'DESCRIPTOR' : _MATRIX,
  '__module__' : 'qm.pb.general_messages_pb2'
  # @@protoc_insertion_point(class_scope:qm.grpc.general_messages.Matrix)
  })
_sym_db.RegisterMessage(Matrix)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



import datetime
import json
import sys
import traceback
import types
from typing import Optional

from qm import _Program, version
from qm.QuaNodeVisitor import QuaNodeVisitor
from qm.QuaSerializingVisitor import QuaSerializingVisitor
from qm.pb.inc_qua_config_pb2 import QuaConfig
from qm.pb.inc_qua_pb2 import QuaProgram
from qm.program import load_config
from qm.program.ConfigBuilder import convert_msg_to_config


def generate_qua_script(prog: _Program, config: Optional[dict] = None) -> str:
    if prog.is_in_scope():
        raise RuntimeError("Can not generate script inside the qua program scope")

    proto_config = None
    if config is not None:
        proto_config = load_config(config)

    proto_prog = prog.build(QuaConfig())
    return _generate_qua_script_pb(proto_prog, proto_config, config)


def _generate_qua_script_pb(proto_prog, proto_config, original_config):
    pretty_original_config = None
    if original_config is not None:
        pretty_original_config = json.dumps(original_config, sort_keys=False, indent=4)
    pretty_proto_config = None
    if proto_config is not None:
        normalized_config = convert_msg_to_config(proto_config)
        pretty_proto_config = json.dumps(normalized_config, sort_keys=False, indent=4)
    extra_info = ""
    serialized_program = ""
    try:
        visitor = QuaSerializingVisitor()
        visitor.visit(proto_prog)
        serialized_program = visitor.out()

        extra_info = _validate_program(proto_prog, serialized_program)
    except Exception as e:
        trace = traceback.format_exception(*sys.exc_info())
        extra_info = extra_info + _error_string(e, trace)

    return f"""
# Single QUA script generated at {datetime.datetime.now()}
# QUA library version: {version.__version__}

{serialized_program}
{extra_info if extra_info else ""}
config = {pretty_original_config}

loaded_config = {pretty_proto_config}

"""


def _validate_program(old_prog, serialized_program: str) -> Optional[str]:
    generated_mod = types.ModuleType("gen")
    exec(serialized_program, generated_mod.__dict__)
    new_prog = generated_mod.prog.build(QuaConfig())

    new_prog_str = _program_string(new_prog)
    old_prog_str = _program_string(old_prog)

    if new_prog_str != old_prog_str:
        return f"""

####     SERIALIZATION WAS NOT COMPLETE     ####
#
#  Original   {old_prog_str}
#  Serialized {new_prog_str}
#
################################################

        """

    return None


def _error_string(e: Exception, trace) -> str:
    return f"""

    ####     SERIALIZATION VALIDATION ERROR     ####
    #
    #  {str(e)}
    #
    # Trace:
    #   {str(trace)}
    #
    ################################################

            """


def _program_string(prog) -> str:
    """
    Will create a canonized string representation of the program
    """
    copy = QuaProgram()
    copy.CopyFrom(prog)
    strip_location_visitor = _StripLocationVisitor()
    strip_location_visitor.visit(copy)
    string = str(copy)
    string = string.replace("\n", "")
    return string


class _StripLocationVisitor(QuaNodeVisitor):
    """
    Go over all nodes and if they have a location property, we strip it
    """

    def _default_enter(self, node):
        if hasattr(node, "loc"):
            node.loc = "stripped"
        return True

    @staticmethod
    def strip(node):
        _StripLocationVisitor().visit(node)
